/******************** files.c ******************
	contains the routines to handle the internal file
	representations.
	e.g., inserting and deleting portions of the file,
	memory menagement, undo routines.
	all are not system dependent.
************************************************/

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
	/* for Malloc amd Mfree */

#include "types.h"
#include "header.h"
#include "files.h"
#include "sysdep.h"
#include "menu.h"
#include "blocks.h"

#define max(x,y) ((x)>(y))?(x):(y)

#define GAPSIZE 1024l
#define GAPDIST (16*1024l)

/**************************************************/ 
/*	Data types and functions, handling the files. */
/**************************************************/

filetyp files[MAXSLOTS]; 
	/* contains all file infromation */
char *ramstart,*ramend,*ramfree; 
	/* start and end of RAM,start of free ram */
int stopped;

char path[128]="";
	/* default path */

undotyp undo;
extern int topslot;

/******** handle RAM **************/

int memory (int slot, long size)
/************ memory
	expand the memory of the slot by size. return, if that was
	possible.
*************/
{	int i;
	filetyp *ft;
	char *end=files[slot].end;
	/* see, if there is enough space left */
	if (ramfree+size>ramend) return 0;
	/* move memory up or down */
	if (ramfree>end)
		memmove(end+size,end,ramfree-end);
	/* change file pointers */
	for (i=slot+1; i<MAXSLOTS; i++)
	{	ft=files+i;
		ft->start+=size;
		ft->end+=size;
		ft->fend+=size;
		ft->topl+=size;
		ft->cursor.line+=size;
		ft->cursor.adr+=size;
		if (ft->gap.open)
		{	ft->gap.start+=size;
			ft->gap.end+=size;
		}
	}
	/* change blocks */
	if (block.slot>=0)
	{	if (block.start>=end) block.start+=size;
		if (block.end>=end) block.end+=size;
	}
	/* change markers */
	for (i=0; i<MAXMARKER; i++)
		if ((marker[i].slot>=0)&&(marker[i].line>=end))
			marker[i].line+=size;
	/* correct undo buffer */
	if (undo.slot>=0)
	{	if (undo.undo>=end) undo.undo+=size;
		if (ramfree+size>undo.start) undo.slot=-1;
	}
	/* expand file slot */
	files[slot].end+=size;
	/* correct free ram pointer */
	ramfree+=size;
	return 1;
}

int init_ram (void)
/*************** init_ram
	initialize the RAM pointers ramtart, ramfree, ramend.
	reserves all memory besides RESERVE bytes for system.
	Use mfree and malloc in ANSI style and guess the free memory!
****************/
{	int i;
	long ram;
	/* how much RAM is there? */
	ram=memsize;
	/* to little? */
	if (ram<FRESERVE) return 1;
	/* allocate it */
	ramfree=ramstart=(char *)malloc(ram);
	ramend=ramstart+ram;
	/* set file pointers for empty files */
	for (i=0; i<MAXSLOTS; i++) 
	{	files[i].cursor.x=files[i].cursor.y=files[i].flags.used=0;
		files[i].start=files[i].end=files[i].fend=files[i].topl=
			files[i].cursor.line=files[i].cursor.adr=ramfree;
	}
	/* no block defined, of course */
	block.start=block.end=0; block.slot=-1;
	/* and no marker */
	for (i=0; i<MAXMARKER; i++) marker[i].slot=-1;
	/* undo buffer invalid */
	undo.slot=-1;
	topslot=-1;
	return (!ramstart);
}

void exit_ram (void)
/*********** exit_ram
	get rid of all ram,
************/
{	free(ramstart);
}

/************* insert and delete blocks of file **********/

int finsert (int slot, char *p, size_t size)
/**************** finsert
	insert some bytes at p. adjust all pointers, blocks and
	markers, and the gap. 
****************/
{	int i;
	filetyp *ft;
	size_t gapsize;
	ft=files+slot;
	if (size<=0) return 0;
	if (!ft->gap.open || p>=ft->gap.end) /* gap not important */
	{	/* make space for that operation first */
		if (size+FRESERVE/4>(ft->end-ft->fend))
			if (!memory(slot,((size+FRESERVE-(ft->end-ft->fend))/4)*4))
			{	error("Out of memory!");
				return 0;
			}
		/* make a gap of size */
		memmove(p+size,p,fend(ft)-p);
		/* adjust all pointers */
		fend(ft)+=size;
		if (ft->cursor.line>p) 
		{	ft->cursor.line+=size; ft->cursor.adr+=size; }
		else if (ft->cursor.adr>p)
		{	ft->cursor.adr=ft->cursor.line;
			ft->cursor.x=0;
		}
		if (ft->topl>p) ft->topl+=size;
		if (block.slot==slot)
		{	if (block.start>=p) block.start+=size;
			if (block.end>p) block.end+=size;
		}
		for (i=0; i<MAXMARKER; i++)
			if (marker[i].slot==slot)
				if (marker[i].line>=p) marker[i].line+=size;
		if (undo.slot==slot)
			if (undo.undo>=p) undo.undo+=size;
	}
	else /* gap is important */
	{	gapsize=ft->gap.end-ft->gap.start;
		if (gapsize<size) /* gap is to short, widen it */
		{	if (!finsert(slot,ft->gap.end,((size-gapsize+GAPSIZE)/4)*4))
			{	/* short of memory */
				return 0;
			}
			ft->gap.end+=((size-gapsize+GAPSIZE)/4)*4;
		}
		memmove(p+size,p,ft->gap.start-p);
		/* adjust all pointers */
		if (ft->cursor.line>p)  
		{	if (ft->cursor.line<=ft->gap.start) 
			{	ft->cursor.line+=size; ft->cursor.adr+=size; 
			}
		}
		else if (ft->cursor.adr>p)
		{	ft->cursor.adr=ft->cursor.line;
			ft->cursor.x=0;
		}
		if (ft->topl>p && ft->topl<ft->gap.start) ft->topl+=size;
		if (block.slot==slot)
		{	if (block.start>=p && block.start<ft->gap.start) 
				block.start+=size;
			if (block.end>p && block.end<ft->gap.start) 
				block.end+=size;
		}
		for (i=0; i<MAXMARKER; i++)
			if (marker[i].slot==slot)
				if (marker[i].line>=p && marker[i].line<ft->gap.start) 
					marker[i].line+=size;
		if (undo.slot==slot)
			if (undo.undo>=p && undo.undo<ft->gap.start) 
				undo.undo+=size;
		ft->gap.start+=size;
		if (ft->gap.start==ft->gap.end) ft->gap.open=0;
	}
	return 1;
}

void fdelete (int slot, char *p, size_t size)
/************ fdelete
	delete size bytes from file.
	adjust the pointers, if necessary, blocks and markers.
	adjust the gap.
	also the cursor position may no longer be valid,
	adjust that also.
************/
{	int i;
	filetyp *ft;
	size_t gapsize,delsize;
	ft=files+slot;
	if (size<=0) return;
	if (!ft->gap.open || p>=ft->gap.start) /* gap does not matter */
	{	/* delete the bytes */
		if (size>(ft->fend-p)) size=ft->fend-p;
		if (size==0) return;
		memmove(p,p+size,fend(ft)-(p+size));
		/* adjust pointers */
		fend(ft)-=size;
		if (ft->topl>=p)
			if (ft->topl<p+size) ft->topl=p;
			else ft->topl-=size;
		ft->topl=prev_line(ft->topl+1,ft);
		if (ft->cursor.line>p)
			if (ft->cursor.line<p+size) 
			{	ft->cursor.line=p;
				ft->cursor.line=
					prev_line(ft->cursor.line+1,ft);
				ft->cursor.adr=ft->cursor.line; ft->cursor.x=0;
			}
			else
			{	ft->cursor.line-=size;
				ft->cursor.adr-=size;
			}
		else if (ft->cursor.adr>p)
		{	ft->cursor.adr=ft->cursor.line; ft->cursor.x=0;
		}
		if (block.slot==slot)
		{	if (block.start>=p+size) 
			{	block.start-=size;
				if (block.start<p) block.start=p;
			}
			if (block.end>=p+size) 
			{	block.end-=size;
				if (block.end<p) block.end=p;
			}
		}
		/* adjust the marker */
		for (i=0; i<MAXMARKER; i++)
			if (marker[i].slot==slot)
			{	if (marker[i].line>=p+size) marker[i].line-=size;
				else if (marker[i].line>=p) marker[i].line=p;
			}
		/* adjust the gap */
		if (ft->gap.open)
		{	if (ft->gap.end>p) ft->gap.end=p;
			if (ft->gap.end==ft->gap.start) ft->gap.open=0;
		}
		/* see if we have used to much memory */
		if ((ft->end-ft->fend)>2*FRESERVE)
		{	memory(slot,-(((ft->end-ft->fend)-FRESERVE)/4)*4);
		}
	}
	else /* gap does matter */
	{	if (p+size>ft->gap.start) /* delete includes gap */
		{	ft->gap.open=0;
			fdelete(slot,p,size);
		}
		else /* delete a part in front of gap */
		{	memmove(p,p+size,ft->gap.start-(p+size));
			/* adjust pointers */
			ft->gap.start-=size;
			if (ft->topl>=p && ft->topl<ft->gap.start)
				if (ft->topl<p+size) ft->topl=p;
				else ft->topl-=size;
			ft->topl=prev_line(ft->topl+1,ft);
			if (ft->cursor.line>p) 
				if (ft->cursor.line<ft->gap.start)
				{	if (ft->cursor.line<p+size) 
					{	ft->cursor.line=p;
						ft->cursor.line=
							prev_line(ft->cursor.line+1,ft);
						ft->cursor.adr=ft->cursor.line; 
						ft->cursor.x=0;
					}
					else
					{	ft->cursor.line-=size;
						ft->cursor.adr-=size;
					}
				}
			else if (ft->cursor.adr>p && ft->cursor.adr<ft->gap.start)
			{	ft->cursor.adr=ft->cursor.line; ft->cursor.x=0;
			}
			if (block.slot==slot)
			{	if (block.start>=p+size && block.start<ft->gap.start) 
				{	block.start-=size;
					if (block.start<p) block.start=p;
				}
				if (block.end>=p+size && block.end<ft->gap.start) 
				{	block.end-=size;
					if (block.end<p) block.end=p;
				}
			}
			/* adjust the marker */
			for (i=0; i<MAXMARKER; i++)
				if (marker[i].slot==slot)
				{	if (marker[i].line>=p+size) 
					{	if (marker[i].line<ft->gap.start) 
							marker[i].line-=size;
					}
					else if (marker[i].line>=p) 
						marker[i].line=
							prev_line(p+1,ft);
				}
			/* see, if gap is to long now */
			gapsize=ft->gap.end-ft->gap.start;
			if (gapsize>2*GAPSIZE)
			{	delsize=((gapsize-GAPSIZE)/4)*4;
				fdelete(slot,ft->gap.end-delsize,delsize);
			}
		}
	}
}

/************* kill one file ****************/

void freset (int slot)
/************
	Reset a slot to empty.
*************/
{	filetyp *ft;
	int i;
	ft=files+slot;
	ft->flags.used=0;
	ft->gap.open=0;
	if (block.slot==slot) block.slot=-1;
	for (i=0; i<MAXMARKER; i++)
		if (marker[i].slot==slot) marker[i].slot=-1;
	if (undo.slot==slot) undo.slot=-1;
	memory(slot,-(ft->end-ft->start));
	fend(ft)=start(ft);
}

/* move lines up and down in a file */

char *prev_line (char *line, filetyp *ft)
/***** prev_line
	find the previous line.
	take care of the gap!
*****/
{	char *start=ft->start;
	if (ft->gap.open)
	{	if (line>ft->gap.end) start=ft->gap.end;
		else if (line==ft->gap.end) line=ft->gap.start;
	}
	if ((line>=start+1) && (*(line-1)==10)) line--;
	if (line>=ft->fend) line=ft->fend-1;
	while (line>start)
	{	line--;
		if (*line==10) return line+1;
	}
	return start;
}

char *next_line (char *line, filetyp *ft)
/****** next_line  
	find the next line
******/
{	char *old=line,*end=ft->fend;
	while (line<end) 
		if (*line==10)
		{	if (!ft->gap.open || line+1!=ft->gap.start) 
				{ return line+1; }
			else { return ft->gap.end; }
		}
		else line++;
	return old;
}

/* routines to connect files position with cursor positions */

int find_pos (char *line, int pos, char *end, char **res,
	int tabsize)
/********** find_pos
	find the position of the cursor in the line. pos is planned
	position, returns pointer in *res and actual position.
	handles mainly the TABs.
**********/
{	int count=0;
	while ( (*line!=10) && (line<end) && (count<pos) )
	{	if (*line==9)
			if ((count/tabsize+1)*tabsize<=pos)
				count=(count/tabsize+1)*tabsize;
			else
			{	break;
			}
		else count++;
		line++;
	}
	*res=line;
	return count;
}

int position (char *line, char *end, int tabs, char *p)
/********** position
	sees where the pointer p will apear on screen. returns the
	cursor position.
***********/
{	int count=0;
	while ((line<p) && (line<end))
	{	if (*line==9) count=(count/tabs+1)*tabs;
		else count++;
		line++;
	}
	return count;
}

/****************************************************/
/* Routines, to load and save a file                */
/****************************************************/

void load_it (char *name, int slot, int flag)
/************* load_it
	load a file from disk. name is name. slot is the slot
	to which the file is to be loaded.
	open a window for that file.
	flag==0 means that default window elements should be used,
	else window elements are given in the slot. 
**************/
{	int newfile,mode;
	char name1[256];
	char *p;
	size_t size,fsize;
	filetyp *ft;

	/* find the file */
	if (search_file(name,name1,&fsize,&mode)) newfile=0;
	else newfile=1;

	/* see, if we have enough RAM */
	if (newfile) size=FRESERVE;
	else size=((fsize+FRESERVE)/2)*2;
	if (size>freesize())
	{	error("Out of memory!|(use -sNNN in command line)");
		return;
	}

	/* make space for file */
	memory(slot,size);
	ft=files+slot;
	ft->mode=0;
	p=start(ft);
	if (!newfile) finsert(slot,p,fsize);
	ft->flags.used=1;
	ft->topl=p; /* first line on display */
	ft->left=0; /* all from leftmost column */
	if (!flag) 
	{	ft->tabs=params.tabsize;
		ft->linelength=params.linelength;
		ft->flags.insert=params.insert;
		ft->flags.indent=params.indent;
		ft->flags.wordwrap=params.wordwrap;
	}
	ft->flags.updated=0;
	ft->flags.info=0;
	ft->cursor.line=p;
	ft->cursor.adr=p;
	ft->cursor.x=0;
	ft->cursor.y=0;
	ft->gap.open=0;

	/* load it */
	if (!newfile) 
		if (!memload(name,p,fsize)) 
		{	freset(slot);
			return;
		}
	else (ft->mode=mode);
	strcpy(ft->name,name);

	open_fwindow (slot,flag);
}

void do_load (char *name)
/********** do_load
	is called from the menu. and calls load_it.
	finds an empty slot.
***********/
{	int slot;

	/* find an empty slot */
	for (slot=0; slot<MAXSLOTS; slot++)
		if (!files[slot].flags.used) break;
	if (slot>=MAXSLOTS)
	{	error("No more files, please!");
		return;
	}

	load_it(name,slot,0);
}

void load_file (void)
/********** load_file
	selet a file and load it into memory.
***********/
{	int slot;
	static char string[256]="";
	
	/* find an empty slot */
	for (slot=0; slot<MAXSLOTS; slot++)
		if (!files[slot].flags.used) break;
	if (slot>=MAXSLOTS)
	{	error("No more files, please!");
		return;
	}
	
	select_file ("Load : ",string);
	
	/* load it, if possible */
	if (string[0]) load_it(string,slot,0);
}

int save_it (int slot)
/********** save_it
	saves a file to disk.
**********/
{	int res;
	filetyp *ft=files+slot;
	if (params.dup) make_dup(files[slot].name);
	close_gap(slot);
	res=memsave(ft->name,start(ft),fend(ft)-start(ft));
	if (res) set_mode(ft->name,ft->mode);
	return res;
}

/************ undo routines ***********/
/* an undo buffer is supported at the end of the ram.
   works as a sort of clipboard.
*/

void to_undo (char *p, size_t size, int slot)
{	if (ramend-ramfree<size)
	{	undo.slot=-1; return; }
	memmove(ramend-size,p,size);
	undo.slot=slot;
	undo.start=ramend-size;
	undo.end=ramend;
	undo.undo=p;
}

void from_undo (void)
/********* from_undo
	copy the undo buffer, if valid, back to the file.
**********/
{	size_t size;
	char *p;
	if (undo.slot<0)
	{	error("Undo not possible!"); return; }
	size=undo.end-undo.start;
 	p=undo.undo;
	if (!finsert(undo.slot,undo.undo,size)) return;
	if (undo.slot>=0)
	{	memmove(p,undo.start,size);
		fredraw(undo.slot);
	}
	else
		fdelete(undo.slot,undo.undo,size);
}

int xequal (char *a, char *b)
{	while (1)
	{	if (toupper(*a++)!=toupper(*b++)) return 0;
		if (*a==0)
			if (*b==0) return 1; else return 0;
		if (*b==0) return 0;
	}
}			

int name_equal (char *a, char *b)
{	char *a1,*b1;
	a1=a+strlen(a);
	while (a1>a && *(a1-1)!='\\' && *(a1-1)!=':') a1--;
	b1=b+strlen(b);
	while (b1>b && *(b1-1)!='\\' && *(b1-1)!=':') b1--;
	if (a1==a || b1==b || *(a1-1)==':' || *(b1-1)==':')
		return (xequal(a1,b1));
	return (xequal(a,b));
}

void get_arguments (int n, char *a[])
/********** get_arguments
	load files in argument line 
***********/
{	int i,j,number=0,numberfound=0;
	for (i=1; i<n; i++) 
	{	if (a[i][0]=='+')
		{	number=atoi(a[i]+1);
			numberfound=1;
			continue;
		}
		for (j=0; j<MAXSLOTS; j++)
			if (files[j].flags.used && name_equal(a[i],files[j].name))
				goto noload;
		do_load(a[i]);
		if (numberfound)
		{	goto_line(number);
			numberfound=0;
		}
		noload : ;
	}
}

/***************** gap routines ******************/
/* gaps are viewed as part of the file. however, finsert
   and fdelete, as well as prev_line and next_line must
   take care of the gap.
*/

void close_gap (int slot)
/****** close_gap
   close the gap, if defined.
******/
{	filetyp *ft=files+slot;
	if (ft->gap.open)
	{	fdelete(slot,ft->gap.start,ft->gap.end-ft->gap.start);
		ft->gap.open=0;
	}
}

void open_gap (char *p, int slot)
/****** open_gap
	open a gap in the slot, far enough to be visible.
*****/
{	filetyp *ft=files+slot;
	if (ft->gap.open) return;
	p+=GAPDIST;
	if (p>ft->fend) return;
	p=prev_line(p+2,ft);
	if (!finsert(slot,p,GAPSIZE)) return;
	ft->gap.start=p; ft->gap.end=p+GAPSIZE;
	ft->gap.open=1;
}

void update_gap (char *p, int slot)
/******  update_gap
	make a good gap in the slot.
*****/
{	filetyp *ft=files+slot;
	char *q;
	size_t gapsize;
	if (!ft->gap.open) { open_gap(p,slot); 
		return; }
	if (p+GAPDIST>ft->fend) /* no gap needed */
		{	close_gap(slot);
			return;
		}
	if (ft->gap.start-p>2*GAPDIST) /* gap to far away */
	{	p=prev_line(p+GAPDIST+2,ft);
		gapsize=ft->gap.end-ft->gap.start;
		if (finsert(slot,p,gapsize))
		{	ft->gap.start=p; ft->gap.end=p+gapsize;
			ft->gap.open=1;
		}
		else
		{	error("Some strange error|with the gap");
			return;
		}
		return;
	}
	if (ft->gap.start<p+GAPDIST/2) /* gap to close */
	{	q=ft->gap.start; gapsize=ft->gap.end-ft->gap.start;
		ft->gap.start=ft->gap.end=prev_line(p+GAPDIST+2,ft);
		fdelete(slot,q,gapsize);
		return;
	}
}
